// ************************************************************************************************ 

// Steuerung fr Wechselrichter, 

// ************************************************************************************************ 
// Copyright keins
// Version 
/*
* Chip type           : ATMEGA 168
* Clock frequency     : Internal clock 8 Mhz 

Ein- und Ausgnge
PORT B 0 (14) = LCD Stecker 6 / Anschluss 11 am LCD = DB4 + Taster 1
PORT B 1 (15) = LCD Stecker 3 / Anschluss 12 am LCD = DB5 + Taster 2
PORT B 2 (16) = LCD Stecker 4 / Anschluss 13 am LCD = DB6 + Taster 3
PORT B 3 (17) = LCD Stecker 1 / Anschluss 14 am LCD = DB7 + Taster 4
PORT B 4 (18) = LCD Stecker 9 / Anschluss  4 am LCD = RS
PORT B 5 (19) = LCD Stecker 7 / Anschluss  6 am LCD = E Enable
PORT B 6 ( 9) = LCD Stecker 8 / Common Taster
PORT B 7 (10) = PB3 an PWM Generator: START Wechselrichter

PORT C 0 (23) = X3/9
PORT C 1 (24) = X3/8
PORT C 2 (25) = X3/7 
PORT C 3 (26) = X3/6 
PORT C 4 (27) = X3/5 	
PORT C 5 (28) = X3/4 

PORT D 0 ( 2) = X3/3 und PD2 an PWM Generator 
PORT D 1 ( 3) = X3/2 und PD3 an PWM Generator 

PORT D 2 ( 4) = Sollwert PC5 an PWM Generator 
PORT D 3 ( 5) = Sollwert PC4 an PWM Generator
PORT D 4 ( 6) = Sollwert PC3 an PWM Generator
PORT D 5 (11) = Sollwert PC2 an PWM Generator
PORT D 6 (12) = Sollwert PC1 an PWM Generator
PORT D 7 (13) = Sollwert PC0 an PWM Generator
*/


// 
// 

// ************************************************************************************************ 
// Include Dateien laden
// ************************************************************************************************ 


#include	<avr\io.h>	//AVR Register und Konstantendefinitionen
#include	<avr\interrupt.h> //AVR Interrupt Vektoren
#include	<stdlib.h>
#include	<inttypes.h>
#include "lcd-routines.h"
#include <util/delay.h>
              


unsigned char taster0=0;
unsigned char taster1=0;
unsigned char taster2=0;
unsigned char taster3=0;

unsigned int sollwert = 0;  
unsigned char u_sollwert = 0; 
unsigned char ausgang = 0; 
unsigned char run = 0;  









// *ADC initialisieren*****************************************************************************
// in "ADMUX" muss die Referenzspannung zum Verleich mit der zu messenden Spannung
// eingestellt werden. Hier eingestellt ist AVCC als  Referenzspannung.
// Wenn AVCC auf VCC liegt, ist diese Spannung nicht bermig 
// przise, aber man hat einen greren Eingangsspannungsbereich von 0-5V
// In der Regel ist das vllig ausreichend.
// in "ADCSRA" wird mit ADPS2 bis ADPS0 die Wandlungfrequenz eingestellt.
// Diese soll lt. Datenblatt zwischen 50 KHz und 200KHz liegen. Hier ist 125KHz eingestellt.
// ************************************************************************************************ 
 
void ADC_Init(void) {
 
  uint16_t result;
 
  ADMUX = (1<<REFS0);      			// AVCC als  Referenzspannung nutzen
  ADCSRA = (1<<ADPS2) | (1<<ADPS1);     	// Frequenzvorteiler 64 ==> 8.000.000/64 = 125.000Hz
  ADCSRA |= (1<<ADEN);                  	// ADC aktivieren
 
  /* nach Aktivieren des ADC wird ein "Dummy-Readout" empfohlen, man liest
     also einen Wert und verwirft diesen, um den ADC "warmlaufen zu lassen" */
 
  ADCSRA |= (1<<ADSC);                  // eine ADC-Wandlung 
  while (ADCSRA & (1<<ADSC) );          // auf Abschluss der Konvertierung warten
  /* ADCW muss einmal gelesen werden, sonst wird Ergebnis der nchsten
     Wandlung nicht bernommen. */
  result = ADCW;
}
// ************************************************************************************************ 


// *ADC Einzelmessung ****************************************************************************** 
uint16_t ADC_Read( uint8_t channel )
{
  // Kanal waehlen, ohne andere Bits zu beeinfluen
  ADMUX = (ADMUX & ~(0x1F)) | (channel & 0x1F);
  ADCSRA |= (1<<ADSC);            // eine Wandlung "single conversion"
  while (ADCSRA & (1<<ADSC) )     // auf Abschluss der Konvertierung warten
    ;
  return ADCW;                    // ADC auslesen und zurckgeben
}
// ********************************************************************************** 



// *******Abfrage Taster  **************************************** 
// Durch die Bettigung der Taster wird das Signal an PB6 auf 0 gezogen
// ******************************************************************* 
unsigned char pushbutton(uint8_t chanel)
{
uint8_t hilf=0;

PORTB |= _BV(6);	// Pull up Widerstand Pin 6 einschalten
			// muss bei jedem Aufruf geschehen, da die
			// LCD Routinen auf den ganzen PortB schreiben
PORTB |= _BV(0);
PORTB |= _BV(1);
PORTB |= _BV(2);
PORTB |= _BV(3);

PORTB &=~_BV(chanel);	// Gegenstck zum abzufragenden Taster auf 0 setzen
_delay_us(42);		// Wartezeit bis der Signalwechsel auch am Pin erkannt wird
if ( !(PINB & (1<<6)) ) hilf=0; else hilf=1; // wenn PB6 =0 dann Taser bettigt
return(hilf);
}




// ********************************************************************************** 
//  Display integer 0-9999 
//  Durch geschicktes multiplizieren und dividieren wird die Integer-Zahl
//  in ihre einzelnen Stellen zerlegt und diese dann am LCD ausgegeben.
//  Es wird nicht abgefragt, ob die Zahl grer als 9999 ist!
//  Whre dies der Fall, wird ein falsches Zeichen am LCD angezeigt.
// ********************************************************************************** 
void print4(unsigned int x)
{
  unsigned int y;
    y=x/1000;lcd_data(y+0x30);x-=(y*1000);
    y=x/100;lcd_data(y+0x30);x-=(y*100);
    y=x/10;lcd_data(y+0x30);x-=(y*10);
    lcd_data(x+0x30);
}


// ********************************************************************************** 
//  spiegelt die Bits in einem Byte
// ********************************************************************************** 
char bytemirror (char n)
{
  n = ((n >> 1) & 0x55) | ((n << 1) & 0xaa); 
  n = ((n >> 2) & 0x33) | ((n << 2) & 0xcc); 
  n = ((n >> 4) & 0x0f) | ((n << 4) & 0xf0); 
  return n;
}


char bitsreverse (uint8_t n)
{
    uint8_t out = 0;
    uint8_t loop;
 
    for (loop = 0; loop < 8; loop++)
    {
       out <<= 1;
 
       if (n & 0x01)
       {
           out |= 0x01;
       }
 
       n  >>= 1;
    }
    return (out);
}

 




// ******************************************************************* 
//                       Timerinterrupt 0  
//
// Zeit fr den Aufruf dieser Routine:
// Zeit = 1 / (1.000.000/(1024*256) = 0,26214sec
// ******************************************************************* 
ISR (TIMER0_OVF_vect)             // signal handler for tcnt0 overflow interrupt
{




TCNT0 = 256-256;  	// Die 256 werden von 256 abgezogen, 
					// da der Timerroutine bei einem berlauf aufgerufen wird
}



// ******************************************************************* 
//                       Timerinterrupt 1  
//
// Zeit fr den Aufruf dieser Routine:
// Zeit = 1 / (1.000.000/(1024*488) = 0,499sec
// ******************************************************************* 
ISR (TIMER1_OVF_vect) 	// signal handler for tcnt1 overflow interrupt
{

//if (taster0==0) u_sollwert++;
//if (taster2==0) u_sollwert--;



TCNT1 = 65536-3904;  	// Die 488 werden von 65536 abgezogen, 
				// da der Timerroutine bei einem berlauf aufgerufen wird
}







// ************************************************************************************************ 
//                       Initialisierung          
// ************************************************************************************************ 
void initial(void)
{    
// **** Einrichten von Ein- und Ausgngen ******************************************************* 
DDRB |= _BV(7);      	// PortB7 Ausgange 
DDRB &=~_BV(6);			// B6 Eingang fr Tasterabfrage
DDRC = 0x00;      	// PortC Eingnge 
DDRD = 0xfc;



// **** Einrichten von Timer 0 *************** 
TCCR0B = (1<<CS02)|(1<<CS00);	// set pre-scaler to (f_clk/1024)  
TIMSK0 = (1<<TOIE0); 		// enable TCNT1 overflow
TCNT0 = 0;  			// Laden des Zeitwertes

// **** Einrichten von Timer 1 *************** 
TCCR1B = (1<<CS02)|(1<<CS00);	// set pre-scaler to (f_clk/1024)  
TIMSK1 = (1<<TOIE1); 		// enable TCNT1 overflow
TCNT1 = 65536-488;  		// Laden des Zeitwertes
sei();				// enable interrupts


}
// ************************************************************************************************ 














// ********************************************************************************** 
// Hauptprogramm
// ********************************************************************************** 

int main(void)
{


 // ******** Initialisierung ********************************************************

initial();	// Initialisierung Ports etc.
ADC_Init(); // Initialisierung Analogeingang
lcd_init(); // Initialisierung des LCD

// ******** Hauptschleife ***********************************************************	
    
  while(1)
  {




// *Taster Abfrage ************************************************************************ 

taster0=pushbutton(0);	// Abfrage erster Taster
taster1=pushbutton(1);	// Abfrage zweiter Taster
taster2=pushbutton(2);	// Abfrage dritter Taster
taster3=pushbutton(3);	// Abfrage vierter Taster



// **** Sollwert berechnen und ausgeben ********************************************* 

//sollwert = ADC_Read(0); // einlesen vom Analogeingang 0 und speichern in "variable"


sollwert = ADC_Read(0); // einlesen vom Analogeingang 0 und speichern in "variable"



u_sollwert=sollwert/16;	// Spannungssollwert skalieren

ausgang= (bytemirror(u_sollwert));
//ausgang= (ausgang);			// die ersten beiden Bits immer auf 1 fr Pullup der Eingnge

PORTD=ausgang;

if (taster3==0) run=1;
if (taster1==0) run=0;


if (run) PORTB |= _BV(7); else PORTB &=~_BV(7);





// *LCD Anzeige ****************************************************************************** 

  lcd_data( 'S' );	// Ausgabe einzelner Zeichen am LCD
  lcd_data( 'W' );
  lcd_data( '=' );
  print4(sollwert);	// Ausgabe einer vierstelligen Integer-Zahl am LCD

  lcd_data( ' ' );	// Ausgabe einzelner Zeichen am LCD
  lcd_data( 'u' );	// Ausgabe einzelner Zeichen am LCD
  lcd_data( 'S' );	// Ausgabe einzelner Zeichen am LCD
  lcd_data( 'W' );
  lcd_data( '=' );



  print4(u_sollwert);	// Ausgabe einer vierstelligen Integer-Zahl am LCD



  lcd_setcursor( 0, 2 );// den Curser am LCD wieder auf Anfang fr den nchsten Programmdurchlauf

if ( run )  lcd_data( '1' );   else lcd_data( '0' );

lcd_data( ' ' );	// Ausgabe einzelner Zeichen am LCD
lcd_data( 'P' );	// Ausgabe einzelner Zeichen am LCD
lcd_data( 'D' );	// Ausgabe einzelner Zeichen am LCD
lcd_data( '=' );

if ( PIND & (1<<7) )  lcd_data( '1' );   else lcd_data( '0' );
if ( PIND & (1<<6) )  lcd_data( '1' );   else lcd_data( '0' );
if ( PIND & (1<<5) )  lcd_data( '1' );   else lcd_data( '0' );
if ( PIND & (1<<4) )  lcd_data( '1' );   else lcd_data( '0' );
if ( PIND & (1<<3) )  lcd_data( '1' );   else lcd_data( '0' );
if ( PIND & (1<<2) )  lcd_data( '1' );   else lcd_data( '0' );
if ( PIND & (1<<1) )  lcd_data( '1' );   else lcd_data( '0' );
if ( PIND & (1<<0) )  lcd_data( '1' );   else lcd_data( '0' );

  lcd_setcursor( 0, 1 );// den Curser am LCD wieder auf Anfang fr den nchsten Programmdurchlauf

// ************************************************************************************************ 




// ************************************************************************************************ 
  } // Ende while(1)Endlosschleife
  return 0;
}// Ende Hauptprogramm
